/* $Id: CalcWavelet.java,v 1.4 2010/12/14 08:34:05 mit Exp $ */
/* CalcWavelet.java v0.2 */

public class CalcWavelet implements Runnable {
    private final int Odiv = 24;
    private final double Sigma = 4.0;

    private int Cnt;
    private int Start;
    private int Octn;
    private int Len;
    private int[] Input;
    private Gabor Gb;
    private Thread thr;

    private static boolean RunSW;
    private static int ViewX = 512, ViewY;
    private static double MaxValue;
    private static double[] Wlet;

    public CalcWavelet(int[] data, int cnt) {
	Input = data;
	Cnt = cnt;
    }
    public void Start(int start, int octn, int len) {
        RunSW = true;
	this.Start = start;
	this.Octn = octn;
	this.Len = len;
 
	MaxValue = 0;
	ViewY = Octn*Odiv;
	if (ViewY > 512)
	    ViewY = 512;

        thr = new Thread(this);
        thr.start();
    }
    public void Stop() {
	if (thr != null)
            thr = null;
    }
    public void run() {
	int stime = 10;

	calc_wavelet();
	try {
	    while(isRun())
		thr.sleep(stime);
	} catch(Exception e) {
	    System.out.println("Run error."+e);
	}

	Stop();
    }
    public void calc_wavelet() {
	double d = Sigma*(0x01 << Octn);
	int offset = (int)Math.sqrt(-2.0*d*d*Math.log(0.01));
	int worksz = Len+offset*2;
	int[] work = new int[worksz];
	int ao = Start-offset;
	for (int i = 0; i < worksz; i++)
	    work[i] = ReadData(ao+i);

	Gb = new Gabor(work, worksz);
	Wlet = new double[ViewX*ViewY];
	int l = 0;
	for (int i = 0; i < ViewY; i++) {
	    Gb.SetParameters(Sigma, Math.PI,
			     Math.pow(2.0, (double)i/Odiv));
	    for (int j = 0; j < ViewX; j++) {
		Wlet[l] = Gb.Wavelets((int)(offset+(double)Len*j/ViewX));
		MaxValue = Math.max(Wlet[l], MaxValue);
		l++;
	    }
	}

	RunSW = false;
    }
    public int ReadData(int p) {
        return (p < 0 || p >= Cnt) ? 0 : Input[p];
    }
    //---
    public static boolean isRun() {
	return RunSW;
    }
    public static double[] getData() {
	return Wlet;
    }
    public static double getData(int num) {
	return Wlet[num];
    }
    public static double getMaxValue() {
	return MaxValue;
    }
    public static int getViewX() {
	return ViewX;
    }
    public static int getViewY() {
	return ViewY;
    }
}
