/* $Id: Cwt.java,v 1.30 2010/12/14 08:34:05 mit Exp $ */
/* Cwt.java v0.3 */
/* Continuous Wavelet Transform */

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.io.*;
import java.net.*;
import java.text.DecimalFormat;

public class Cwt extends Applet 
    implements ActionListener, AdjustmentListener, ItemListener, Runnable
{
    private Choice chFile;
    private Choice chOctave;
    private Choice chSample;
    private Button[] btSelect;
    private Button btPwide;
    private Button btMwide;
    private Button btPgain;
    private Button btMgain;
    private Scrollbar scStart;
    private TextField tfA;
    private Canvas can;
    private Panel PU, PL, PR, PD;

    final int WAVE = 0;
    final int FFT_ = 1;
    final int GABO = 2;
    final int Sample[] = {9, 10, 11, 12, 13, 14}; // 512~16384
    final String Select[] = {"Wave","FFT_","Gabor"}; // 0,1,2
    private static final long serialVersionUID = 1L;

    boolean CALW;  // wavelet
    boolean CALF;  // FFT
    int W;
    int H;
    int H2;
    int FNT;
    int Grap;
    int Cnt;
    int Wide;
    int Splngf;
    int AllFile;
    int OctaveN = 6;
    int ViewAnalX = 512;
    int ViewAnalY;
    int AnalyzeLength = 1024;
    int LeftSide;
    int BotomSide;
    int WL, HB;
    int N = (int)Math.pow(2, 10); // 1024
    int N2 = N/2;
    int Wbyf = 2;     // wide fft
    int Splw;         // fft memori
    int ColorN;       //
    double Gain;
    double Gainf = 0.0001; // gain fft
    double Gainc = 1;
    double MaxValue;
    String Dirnm;
    String Error;

    double[] re;      // for fft
    double[] im;      // "
    double[] dat;     // "
    double[] Gauss;   // "
    String[] files;
    Color[] Grad;

    Graphics gi;
    Image ig;
    Font Fnt;
    FontMetrics Fm;
    DecimalFormat df1;
    Thread thr;
    FFT fftN;
    CalcWavelet CW;
    ReadWave RW;

    public void init() {
	String fnt = getParameter("font");
        FNT = (fnt != null) ? Integer.parseInt(fnt) : 12;
        Fnt = new Font("SansSerif", Font.PLAIN, FNT);
        setFont(Fnt);
	Fm = getFontMetrics(Fnt);
	df1 = new DecimalFormat("#0.0#");

        setBackground(Color.lightGray);
        setLayout(new BorderLayout());

        PU = new Panel();
        PU.setLayout(new FlowLayout());
        add("North", PU);

        String fn = getParameter("all_f");
        AllFile = (fn == null) ? 0 : Integer.parseInt(fn);
        if (AllFile == 0)
            Error = "No wave file.";

	Dirnm = getParameter("dirnm");
        Dirnm = (Dirnm == null) ? "" : Dirnm;

        files = new String[AllFile];
        chFile = new Choice();
        for (int i = 0; i < AllFile; i++) {
            files[i] = getParameter("wave"+i);
            chFile.addItem(files[i]);
        }
        chFile.addItemListener(this);
        PU.add(chFile);

        tfA = new TextField(24);
        tfA.setEditable(false);
        PU.add(tfA);

	chSample = new Choice();
        for (int i = 0; i < Sample.length; i++)
            chSample.addItem(""+Sample[i]);
        chSample.addItemListener(this);
	chSample.select(1);
        PU.add(chSample);
	//---

        PL = new Panel();
        PL.setLayout(new GridLayout(5,1,0,0));
        add("West", PL);

	btSelect = new Button[Select.length];
	for (int i = 0; i < Select.length; i++) {
	    btSelect[i] = new Button(Select[i]);
	    btSelect[i].addActionListener(this);
	    PL.add(btSelect[i]);
	}

	PL.add(new Label("Wide J:"));

	chOctave = new Choice();
        for (int i = 0; i < 8; i++)     // 3~10
            chOctave.addItem(""+(i+3));
        chOctave.addItemListener(this);
	chOctave.select("6");
        PL.add(chOctave);
	//---

        PR = new Panel();
        PR.setLayout(new GridLayout(5,1,0,0));
        add("East", PR);

	btPgain = new Button("+");
        btPgain.addActionListener(this);
        PR.add(btPgain);

	btMgain = new Button("-");
        btMgain.addActionListener(this);
        PR.add(btMgain);

	PR.add(new Label());

	btPwide = new Button(">");
        btPwide.addActionListener(this);
        PR.add(btPwide);

	btMwide = new Button("<");
        btMwide.addActionListener(this);
        PR.add(btMwide);
	//---

	PD = new Panel();
        PD.setLayout(new GridLayout(1,1,0,0));
        add("South", PD);

        scStart = new Scrollbar(Scrollbar.HORIZONTAL, 0, 10, 0, 100);
        scStart.addAdjustmentListener(this);
        PD.add(scStart);
	//---

	can = new MyCanvas();
        add(can, "Center");
    }

    public void start() {
	W = getCanvasD().width;
        H = getCanvasD().height;
	LeftSide = Fm.stringWidth("000000");
	BotomSide = FNT*3;
	WL = W-LeftSide;
	HB = H-BotomSide;
	H2 = HB/2;

        can.setSize(W, H);
        ig = createImage(W, H);
        gi = ig.getGraphics();
        gi.setFont(Fnt);

	set_enabled();
	initGain(H2);
	initColor();
	setSample();

	if (AllFile > 0)
            read_file(files[0]);
    }
    public Dimension getCanvasD() {
        Dimension du = PU.getSize();
	Dimension dl = PL.getSize();
	Dimension dr = PR.getSize();
	Dimension dd = PD.getSize();
        Dimension da = getSize();
        int cw = da.width-(dl.width+dr.width);
        int ch = da.height-(du.height+dd.height);
        Dimension can = new Dimension(cw, ch);

        return can;
    }
    public void set_enabled() {
	boolean ng = (Grap != GABO);
	scStart.setEnabled(ng);
	btPwide.setEnabled(ng);
	btMwide.setEnabled(ng);

	boolean fs = (Grap == FFT_);
	chSample.setEnabled(fs);
    }
    public void initColor() {
	int f = 255, z = 255-f, fi;
	ColorN = f*7;
	Grad = new Color[ColorN];
	for (int i = 0; i < f; i++) {
	    fi = f-i;
	    Grad[i] = new Color(f, fi, fi);
	    Grad[i+f] = new Color(f, i, z);
	    Grad[i+f*2] = new Color(fi, f, z);
	    Grad[i+f*3] = new Color(z, f, i);
	    Grad[i+f*4] = new Color(z, fi, f);
	    Grad[i+f*5] = new Color(i, z, f);
	    Grad[i+f*6] = new Color(fi, z, fi);
	}
    }
    //
    public void itemStateChanged(ItemEvent e) {
	if (e.getSource() == chFile)
            read_file(chFile.getSelectedItem());

	if (e.getSource() == chOctave)
	    selectOctave();

	if (e.getSource() == chSample)
	    setSample();

	set_enabled();
	can.repaint();
    }
    public void adjustmentValueChanged(AdjustmentEvent e) {
	if (e.getSource() == scStart)
	    reCalc();

	can.repaint();
    }
    public void actionPerformed(ActionEvent e) {
	if (e.getSource() == btSelect[WAVE])
            Grap = WAVE;
        if (e.getSource() == btSelect[FFT_])
            Grap = FFT_;
        if (e.getSource() == btSelect[GABO])
            Grap = GABO;

	if (e.getSource() == btPgain)
	    setGain(true);
	if (e.getSource() == btMgain)
	    setGain(false);
	if (e.getSource() == btPwide)
	    setWide(true);
	if (e.getSource() == btMwide)
	    setWide(false);

	set_enabled();
	can.repaint();
    }
    //
    public void read_file(String fname) {
	Error = "";
        read_wave(Dirnm+fname);

	if (Error.equals("")) {
	    setAnalyzeLength();
	    setScroll();
	    setSample();
	    reCalc();
	} else {
	    AnalyzeLength = Cnt = 0;
	}
    }
    public boolean isFile() {
	return Error.equals("");
    }
    public void setAnalyzeLength() {
	while (AnalyzeLength > Cnt)
	    AnalyzeLength /= 2;
    }
    public void read_wave(String fn) {
	try {
            URL url = new URL(getDocumentBase(), fn);
            RW = new ReadWave(url);
            Splngf = RW.getSamplf();
            AnalyzeLength = Cnt = RW.getCount();
            Error = RW.getError();
        } catch(Exception e) {
            Error = "File read error.";
        }
    }
    public String dform(double val) {
        return df1.format((float)val);
    }
    public void reCalc() {
	CALW = false;
	CALF = false;
    }
    //
    public void startRun() {
        if (isFile() & !CALW) {
            thr = new Thread(this);
            thr.start();
        }
    }
    public void stop() {
        if (isRun())
            thr = null;
    }
    public boolean isRun() {
	return thr != null;
    }
    public void run() {
        int stime = 10;   // sleep time
        setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

	calc_cwt();
        try {
	    while(CW.isRun()) {
		thr.sleep(stime);
		can.repaint();
	    }
        } catch (InterruptedException e) {
            System.out.println(e);
        }

	stop();
	CALW = true;
        setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        can.repaint();
    }
    //-----

    class MyCanvas extends Canvas {
	private static final long serialVersionUID = 1L;
        public void update(Graphics g) {
            can.paint(g);
        }
        public void paint(Graphics g) {
            paintI();
            g.drawImage(ig, 0, 0, this);
        }
    }
    public void paintI() {
	gi.setColor(Color.black);
        gi.fillRect(0, 0, W, H);

	switch (Grap) {
	case WAVE:
	    disp_wave(gi);
	    break;
	case GABO:
	    disp_wavelet(gi);
	    break;
	case FFT_:
	    disp_fft(gi);
	    break;
	default:
	}

	disp_waku(gi);
	write_text();
    }
    public void initGain(int h2) {
	Gain = (double)h2/Short.MAX_VALUE;
    }
    public void setGain(boolean sw) { //+-
	if (!isFile())
	    return;

	double ms2 = Math.sqrt(2);
	switch (Grap) {
	case WAVE:
	    Gain = sw ? Gain*ms2 : Gain/ms2;
	    break;
	case GABO:
	    Gainc = sw ? Gainc*ms2 : Gainc/ms2;
	    if (Gainc < 1)
		Gainc = 1;
	    break;
	case FFT_:
            Gainf = sw ? Gainf*ms2 : Gainf/ms2;
            break;
	default:
	} 
    }
    public void setWide(boolean sw) { //<>
	if (!isFile())
	    return;

	int min, max;
	switch (Grap) {
	case WAVE:
	    min = 128; //
	    AnalyzeLength = sw ? AnalyzeLength*2 : AnalyzeLength/2;
	    if (AnalyzeLength > Cnt)
		AnalyzeLength = Cnt;
	    else if (AnalyzeLength < min)
		AnalyzeLength = min;
	    else
		CALW = false;
	    setScroll();
	    break;
	case FFT_: // Ψ
	    min = 2;
	    max = 512;
	    Wbyf = sw ? Wbyf*2 : Wbyf/2;
	    if (Wbyf > max)
		Wbyf = max;
	    else if (Wbyf < min)
		Wbyf = min;
	    else
		setScroll();
	    break;
	default:
	}
    }
    public void write_text() {
	StringBuffer str = new StringBuffer();

	if (Error.equals("")) {
	    switch (Grap) {
	    case WAVE:
		str.append(dform(100.0*AnalyzeLength/Cnt)+"% ");
		str.append(getWaveStart()+" (");
		str.append(AnalyzeLength+"/"+Cnt+")");
		break;
	    case FFT_:
	    case GABO:
		str.append(RW.getSamplf()+"[Hz]");
		str.append(dform((double)Cnt/Splngf)+"[sec]:");
		str.append(RW.getLavel());
		break;
	    default:
	    }
	} else
	    str.append(Error);

	tfA.setText(str.toString());
    }
    public void setScroll() {
	scStart.setValues(scStart.getValue(), AnalyzeLength, 0, Cnt);
	scStart.setBlockIncrement(AnalyzeLength);
	scStart.setUnitIncrement(AnalyzeLength/2);
    }
    public int getWaveStart() {
	return scStart.getValue();
    }
    public void selectOctave() {
	OctaveN = Integer.parseInt(chOctave.getSelectedItem());

	CALW = false;
    }
    public void setSample() {
	int n0 = Integer.parseInt(chSample.getSelectedItem()); 
	N = (int)Math.pow(2, n0);
	N2 = N/2;

        fftN = new FFT(N);
	re = new double[N];
        im = new double[N];
        dat = new double[N2];
	init_gauss();

	CALF = false;
    }
    //
    public void calc_cwt() {
	CW = new CalcWavelet(RW.getData(), Cnt);
	CW.Start(getWaveStart(), OctaveN, AnalyzeLength);
	ViewAnalX = CW.getViewX();
	ViewAnalY = CW.getViewY();
    }
    public void disp_wavelet(Graphics g) {
	if (!isFile())
	    return;

	if (!CALW & !isRun())
	    startRun();

	disp_y_line(g);
	disp_time(g);
	colorBar(g);

	MaxValue = CW.getMaxValue();
	double dx = (double)WL/ViewAnalX;
	double dy = (double)HB/ViewAnalY;
	double dc = ColorN/MaxValue;
	int[] x0 = new int[ViewAnalX];
	for(int j = 0; j < ViewAnalX; j++)
	    x0[j] = (int)(dx*j+LeftSide);
	
	int l = 0, y0, max = ColorN-1;
	for(int i = 0; i < ViewAnalY; i++) {
	    y0 = (int)(dy*i);
	    for(int j = 0; j < ViewAnalX; j++) {
		g.setColor(set_color(dc, CW.getData(l++), max));
		g.drawLine(x0[j], y0, x0[j], y0);
	    }
	}
    }
    public void disp_time(Graphics g) {
	double allsec = AnalyzeLength/(double)Splngf; // 
	double tim;
	if (allsec < 0.01)
	    tim = 0.001;
	else if (allsec < 0.02)
	    tim = 0.005;
	else if (allsec < 0.1)
	    tim = 0.01;
	else if (allsec < 0.3)
	    tim = 0.05;
	else if (allsec < 1)
	    tim = 0.1;
	else if (allsec < 3)
	    tim = 0.5;
	else if (allsec < 20)
	    tim = 1;
	else if (allsec < 30)
	    tim = 5;
	else if (allsec < 60)
	    tim = 10;
	else
	    tim = 60;

	double asec = tim/(allsec/WL);
	int maxsec = (int)(allsec/tim);
	if (maxsec > 0) {
	    int ss = H-BotomSide/2; // ʸh
	    int x, l = 5;
	    for (int i = 0; i <= maxsec; i++) {
		x = (int)(asec*i+LeftSide);

		g.setColor(Color.gray);
		g.drawLine(x, HB, x, HB+l);

		g.setColor(Color.lightGray);
		g.drawString(dform(i*tim), x, ss);
	    }

	    if (Grap == WAVE)
		disp_sec(g, ss);
	}
    }
    public void disp_sec(Graphics g, int ss) {
	String s = "[sec]";
	g.setColor(Color.lightGray);
	g.drawString(s, 2, ss);
    }
    public void disp_y_line(Graphics g) {
	double dh = (double)HB/OctaveN;
	int dy, spl = Splngf;

	for (int i = 0; i <= OctaveN; i++) {
	    dy = (int)(dh*i);
	    g.setColor(Color.gray);
	    g.drawLine(0, dy, LeftSide, dy);

	    spl /= 2;
	    g.setColor(Color.orange);
	    g.drawString(""+spl, 0, dy+FNT);
	}
    }
    public Color set_color(double dc, double num, int max) {
	int val = (int)(dc*num*Gainc);
	if (val > max)
	    val = max;
	else if (val < 0) // ͽɤ
	    val = 0;

	return Grad[max-val]; // 顼ȿž
    }
    public void colorBar(Graphics g) {
	int h1 = H-5;
	double dc = (double)ColorN/W;

	for (int i = 0; i < W; i++) {
	    g.setColor(Grad[(int)(i*dc)]);
	    g.drawLine(i, H, i, h1);
	}
    }
    //
    public void disp_wave(Graphics g) {
	if (!isFile())
	    return;

	disp_w_line(g);
	disp_time(g);

	double dx = (double)WL/AnalyzeLength;
	int start = getWaveStart();
	int x, y, dat, max = Integer.MIN_VALUE, min = Integer.MAX_VALUE;

	g.setColor(Color.green);
	for (int i = 0; i < AnalyzeLength; i++) {
	    dat = RW.getData(i+start);
	    x = (int)(dx*i+LeftSide);
	    y = (int)(H2-dat*Gain);
	    g.drawLine(x, y, x, y);

	    max = Math.max(max, dat);
	    min = Math.min(min, dat);
	}

	disp_wave_value(g, min, max);
    }
    public void disp_wave_value(Graphics g, int min, int max) {
        g.setColor(Color.lightGray);
	g.drawString(dform(100.0*max/Short.MAX_VALUE), 1, (int)(H2-max*Gain));
	g.drawString(dform(100.0*min/Short.MIN_VALUE), 1, (int)(H2-min*Gain+FNT));
	g.drawString("[%]", 1, FNT);
    }
    public void disp_w_line(Graphics g) {
	g.setColor(Color.gray);
	g.drawLine(LeftSide, H2, W, H2);
    }
    public void disp_waku(Graphics g) {
	g.setColor(Color.lightGray);
	g.drawLine(LeftSide, 0, LeftSide, HB);
 	g.drawLine(LeftSide, HB, W, HB);
    }
    //
    public void disp_fft(Graphics g) {
	if (!isFile())
	    return;

	if (!CALF)
	    calcFft();

        int x0 = 0, y0 = 0, x1, y1, wmax = N/Wbyf;
	double dw = WL/(double)wmax;

        g.setColor(Color.green);
        for (int i = 0; i < wmax; i++) {
            x1 = (int)(dw*i+LeftSide);
            y1 = (int)(HB-dat[i]*Gainf);
            if (i == 0) {
                x0 = x1;
                y0 = y1;
            }
            g.drawLine(x0, y0, x1, y1);
            x0 = x1;
            y0 = y1;
        }

	disp_fft_memori(g);
    }
    public void disp_fft_memori(Graphics g) {
	Splw = Splngf/Wbyf;
	double ds = (double)WL/Splw;
	int spi, me, hb = HB+1, h = H-FNT;
	for (int i = 0; i < Splw; i += 10) {
	    spi = (int)(ds*i+LeftSide);
	    if (i % 10000 == 0){
		disp_k_memori(g, i, spi, Splw);
		me = 15;
	    } else if (i % 5000 == 0){
		disp_k_memori(g, i, spi, Splw);
		me = 13;
	    } else if (i % 1000 == 0){
		disp_k_memori(g, i, spi, Splw);
		me = 10;
            } else if (Splw < 12000 && i % 500 == 0) {
		me = 7;
            } else if (Splw < 10000 && i % 100 == 0) {
		me = 5;
	    } else if (Splw < 2000) {
		me = 3;
            } else
                continue;

	    g.setColor(Color.lightGray);
            g.drawLine(spi, hb, spi, hb+me);
        }

	String s = "[Hz]";
	g.setColor(Color.pink);
	g.drawString(s, 2, H-FNT);
    }
    public void disp_k_memori(Graphics g, int n, int x, int max) {
	if ((max > 12000 && !(n%5000 == 0)))
	    return;

	g.setColor(Color.pink);
	g.drawString(n/1000+"K", x, H-FNT);
    }
    public void calcFft() {
	gauss();
	fft();

	CALF = true;
    }
    public void init_gauss() {
	Gauss = new double[N];
	double in;
	for (int i = 0; i < N; i++) {
	    in = (2.0*i-N)/N;
	    Gauss[i] = Math.exp(-4.5*in*in);
	}
    }
    public void gauss() { // window
       	int sft = getWaveStart();
        for (int i = 0; i < N; i++) {
	    re[i] = im[i] = RW.getData(i+sft)*Gauss[i];
        }
    }
    public void fft() {   // fft.java
        fftN.fft(re, im);
        for (int i = 0; i < N2; i++)
            dat[i] = Math.sqrt(re[i]*re[i]+im[i]*im[i]);
    }
}